#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <pthread.h>
#include <semaphore.h>

typedef void (*tar_threadstart)(void);

typedef void (*tar_threadcreator)(tar_threadstart);


#define NUM_PHM 1024

struct tar_mem{
  struct tar_mem *next;
  int magic_number;
  size_t size;
  int marked; // Its just a bit, so it could also have been placed in next, magic_number, size, or copy.
  // tar_finalizer finalizer; // Must add. Needed by snd-rt.
} typedef tar_mem_t;


struct tar_heap{
  void *heap;

  tar_mem_t *mem;
  tar_mem_t *atomic_mem;

  // The following four variables can probably be globals.
  tar_mem_t *mem_gc; // mem currently treated by gc
  tar_mem_t *atomic_mem_gc; //mem curently treated by gc
  tar_mem_t *mem_gc_last; // pointer to the last element.
  tar_mem_t *atomic_mem_gc_last; // pointer to the last element.

  tar_mem_t *freelist; // I think this one can be global too. Or maybe not.

  int phms_place;
  void *phms_start[NUM_PHM]; // Should probably be made resizable... (should at least add check)
  void *phms_end[NUM_PHM]; // Should probably be made resizable...

  int num_allocs;

} typedef tar_heap_t;


int tar_find_num_cpus(void);


// Must be called before entering the audio thread.
// Returns false in case it failed.
bool tar_entering_audio_thread(tar_heap_t *heap);


// Must be called before leaving the audio thread.
// Returns true in case the garbage collector is ready to collect
// new garbage.
bool tar_leave_audio_thread(tar_heap_t *tar);

// Only necessary to call before calling tar_run_gc().
void tar_add_root(tar_heap_t *heap,void *start,void *end);

// Must only be called (but is not required to!)
// if tar_leave_audio_thread returned true.
void tar_run_gc(tar_heap_t *heap);


/*
 - atomic_heap_size is the size of the heap holding atomic mem.
   It will not be automatically resized.

 - nonatomic_heap_size is the size of the heaps created when
   calling tar_new_heap(). It will not be automatically resized.

 - roots_size must be large enough to hold the roots.

 - realtime_creator is a function creating a realtime thread.
   The created thread must have the same priority as the audio thread.

 - mark_creator is a function creating a mark thread.
   The created thread does not have to run with
   the same priority as the audio thread, but if using a small
   heap, it might be a good idea to let the mark thread(s)
   run in realtime too, but at a lower priority than the audio thread.
   If the heap is more than large enough, the mark thread(s) should
   probably run with normal priority instead.
*/

void tar_init(size_t atomic_heap_size,
	      size_t nonatomic_heap_size,
	      size_t roots_size,
	      tar_threadcreator realtime_creator,
	      tar_threadcreator mark_creator
	      );


tar_heap_t *tar_new_heap(void);

void tar_delete(tar_heap_t *heap);


// All "alloc" functions may return NULL.

void *tar_alloc(tar_heap_t *heap,size_t size);

void *tar_alloc_atomic(tar_heap_t *heap,size_t size);

// familiar name from hbgc
void *tar_malloc_atomic_uncollectable(size_t size);

// familiar name from hbgc
void *tar_calloc_atomic_uncollectable(size_t size1,size_t size2);

void tar_free_atomic_uncollectable(void *mem);
